/*-
 * Copyright (c) 2007 Michael Telahun Makonnen <mtm@FreeBSD.Org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Copyright (c)2004 The DragonFly Project.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 *   Neither the name of the DragonFly Project nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

/*
 * fn_install.c
 * Installer Function : Install OS Files.
 * $Id: fn_install.c,v 1.74 2006/04/18 19:43:48 joerg Exp $
 */

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/wait.h>

#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <libutil.h>
#include <string.h>
#include <unistd.h>

#define SOURCES_CONF_FILE "/usr/local/share/dfuibe_installer/sources.conf"

#ifdef ENABLE_NLS
#include <libintl.h>
#define _(String) gettext (String)
#else
#define _(String) (String)
#endif

#ifdef SYSTEM_AURA
#include <aura/mem.h>
#include <aura/buffer.h>
#include <aura/fspred.h>
#else
#include "mem.h"
#include "buffer.h"
#include "fspred.h"
#endif

#ifdef SYSTEM_DFUI
#include <dfui/dfui.h>
#include <dfui/system.h>
#else
#include "dfui.h"
#include "system.h"
#endif

#ifdef SYSTEM_LIBINSTALLER
#include <installer/commands.h>
#include <installer/confed.h>
#include <installer/diskutil.h>
#include <installer/functions.h>
#include <installer/uiutil.h>
#else
#include "commands.h"
#include "confed.h"
#include "diskutil.h"
#include "functions.h"
#include "uiutil.h"
#endif

#include "cmdnames.h"
#include "flow.h"
#include "pathnames.h"
#include "fn.h"
#include "dist.h"

#define DIST_MNT		DEFAULT_DESTDIR "dist"
#define DIST_CDROMDEV		"/dev/acd0"
#define INFPATH			DIST_MNT "/cdrom.inf"
#define CDVER			"CD_VERSION"

struct dist {
	char *d_name;
	char *d_distdir;
	char *d_destdir;
};

static int	do_mount(const char *, const char *, const char *, int);

static int
do_mount(const char *dev, const char *mnt, const char *type, int flags)
{
	struct iovec iov[6];

	iov[0].iov_base = "fstype";
	iov[0].iov_len = strlen(iov[0].iov_base) + 1;
	iov[1].iov_base = (void *)type;
	iov[1].iov_len = strlen(iov[1].iov_base) + 1;
	iov[2].iov_base = "fspath";
	iov[2].iov_len = strlen(iov[2].iov_base) + 1;
	iov[3].iov_base = (void *)mnt;
	iov[3].iov_len = strlen(iov[3].iov_base) + 1;
	iov[4].iov_base = "from";
	iov[4].iov_len = strlen(iov[4].iov_base) + 1;
	iov[5].iov_base = (void *)dev;
	iov[5].iov_len = strlen(iov[5].iov_base) + 1;
	return (nmount(iov, 6, flags));
}

/*
 * fn_install_os: actually put FreeBSD on a disk.
 */
void
fn_install_os(struct i_fn_args *a)
{
	struct subpartition *sp;
	struct commands *cmds;
	struct command *cmd;
	struct iovec iovDevfs[4];
	int i, seen_it, prefix, error;
	FILE *sources_conf;
	char line[256];
	char cp_src[64][256];
	char file_path[256];
	char devfsPath[16];
	char distpath[256];
	char dev[MAXPATHLEN];
	char mnt[MAXPATHLEN];
	int lines = 0;
	long mfsbacked_size;
	pid_t fpid;
	int fstatus;
	int cdmounted;

	cdmounted = 0;
	a->result = 0;

	/*
	 * Enable swap partitions
	 */
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	    sp != NULL; sp = subpartition_next(sp)) {
		if (!subpartition_is_swap(sp))
			continue;
		snprintf(dev, sizeof(dev) / sizeof(char), "/dev/%s",
		    subpartition_get_device_name(sp));
		error = swapon(dev);
		i_log(a, "%d = swapon(%s),\t\t errno = %d", error,
		     subpartition_get_device_name(sp), errno);
	}

#if 0
	inform(a->c, "unmount_all_under");
	/*
	 * Unmount anything already mounted on /mnt.
	 */
	unmount_all_under(a, cmds, "%smnt", a->os_root);
#endif

	/*
	 * Create mount points and mount subpartitions on them.
	 * XXX - I assume the root '/' partition is first on the list, so
	 *       subsequent mount points are created on top of it.
	 */
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_next(sp)) {
		if (subpartition_is_swap(sp) || subpartition_is_mfsbacked(sp))
			continue;
	
		snprintf(dev, sizeof(dev) / sizeof(char), "%sdev/%s",
		    a->os_root, subpartition_get_device_name(sp));
		snprintf(mnt, sizeof(mnt) / sizeof(char), "%smnt%s",
		    a->os_root, subpartition_get_mountpoint(sp));
		error = mkdir(mnt,
		    S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
		if (error == -1 && errno != EEXIST) {
			inform(a->c, _("Internal Error (%d) - "
			    "could not create directory %s"), errno, mnt);
			goto dismount;
		}
		i_log(a, "%d = mkdir(%s)\t\t errno = %d", error, mnt, errno);
		i_log(a, "Mounting %s on %s...", dev, mnt);
		if (do_mount(dev, mnt, "ufs", 0) == -1) {
			inform(a->c, _("Internal Error (%d) - unable to "
			    "mount %s on %s"), errno, dev, mnt);
			goto dismount;
		}
	}

	/*
	 * Actually copy files now.
	 */
	struct dist dists[] = {
		{ "base", "base", "/" },
		{ "catpages", "catpages", "/" },
		{ "dict", "dict", "/" },
		{ "doc", "doc", "/" },
		{ "games", "games", "/" },
		{ "info", "info", "/" },
		{ "GENERIC", "kernels", "/boot/" },
#ifdef __amd64__
		{ "lib32", "lib32", "/" },
#endif
		{ "manpages", "manpages", "/" },
/*		{ "ports", "ports", "/usr/" }, */
		{ "proflibs", "proflibs", "/" },
		{ "sbase", "src", "/usr/src/" },
		{ "scddl", "src", "/usr/src/" },
		{ "scompat", "src", "/usr/src/" },
		{ "scontrib", "src", "/usr/src/" },
		{ "scrypto", "src", "/usr/src/" },
		{ "sgnu", "src", "/usr/src/" },
		{ "setc", "src", "/usr/src/" },
		{ "sgames", "src", "/usr/src/" },
		{ "sinclude", "src", "/usr/src/" },
		{ "skrb5", "src", "/usr/src/" },
		{ "slib", "src", "/usr/src/" },
		{ "slibexec", "src", "/usr/src/" },
		{ "srelease", "src", "/usr/src/" },
		{ "sbin", "src", "/usr/src/" },
		{ "ssecure", "src", "/usr/src/" },
		{ "ssbin", "src", "/usr/src/" },
		{ "sshare", "src", "/usr/src/" },
		{ "ssys", "src", "/usr/src/" },
		{ "subin", "src", "/usr/src/" },
		{ "susbin", "src", "/usr/src/" },
		{ "stools", "src", "/usr/src/" },
		{ "srescue", "src", "/usr/src/" },
		{ NULL, NULL, NULL }
	};

	struct dfui_progress *pr;
	struct dist *dp;
	properties infp;
	FILE *ff;
	char fstab[MAXPATHLEN];
	char prompt[256];
	char cdver[256];
	const char *cp;
	int cancelled;
	int skipped;
	int infd;

	/*
	 * Mount release CD-ROM.
	 */
	if (mkdir(DIST_MNT, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) ==
	    -1 && errno != EEXIST) {
		inform(a->c, _("Internal Error (%d) - could not create %s"),
		    errno, DIST_MNT);
		goto dismount;
	}
	i_log(a, "Mounting CD-ROM %s on %s", DIST_CDROMDEV, DIST_MNT);
	if (do_mount(DIST_CDROMDEV, DIST_MNT, "cd9660", MNT_RDONLY) != 0) {
		inform(a->c, _("Internal Error (%d)- could not mount CD-ROM. "
		    "Currently, this installer can only mount the first "
		    "IDE CD-CDROM"), errno);
		goto dismount;
	}
	cdmounted = 1;

	/*
	 * Get the CD version. The path to the directory containing the
	 * distributions is made from this string.
	 */
	infd = open(INFPATH, O_RDONLY);
	if (infd == -1) {
		inform(a->c, _("Internal Error (%d) - Unable to open %s"),
		    errno, INFPATH);
		goto dismount;
	}
	infp = properties_read(infd);
	if (infp == NULL) {
		inform(a->c, _("Internal Error (%d) - Unable to read from %s"),
		    errno, INFPATH);
		goto dismount;
	}
	cp = property_find(infp, CDVER);
	if (cp == NULL) {
		inform(a->c, _("Internal Error - no property: %s"), CDVER);
		goto dismount;
	}
	strncpy(cdver, cp, sizeof(cdver) / sizeof(char));
	properties_free(infp);
	close(infd);
	infp = NULL;
	infd = -1;
	i_log(a, "CD Version String: %s", cdver);

	skipped = 0;
	for (dp = dists; dp->d_name != NULL; dp++) {
retry:
		cancelled = 0;
		snprintf(prompt, sizeof(prompt), _("Extracting %s into %s "
		    "directory"), dp->d_name, dp->d_destdir);
		pr = dfui_progress_new(dfui_info_new(
		_("Installing"),
		prompt,
		""),
		0);

		if (!dfui_be_progress_begin(a->c, pr))
			abort_backend();

		snprintf(distpath, 256, "%s/%s/%s", DIST_MNT, cdver,
		    dp->d_distdir);
		i_log(a, "Extracting %s from %s to %s", dp->d_name, distpath,
		    dp->d_destdir);
		error = distExtractTarball(a, distpath, dp->d_name,
		    dp->d_destdir, pr, &cancelled);
		if (error == 0) {
			if (!dfui_be_progress_end(a->c))
				abort_backend();
			switch (dfui_be_present_dialog(a->c,
			    _("Extraction failed!"),
			    _("Retry|Skip|Cancel"),
			    _("Extraction of %s failed.\n\n"),
			    dp->d_name)) {
			case 1:
				dfui_progress_free(pr);
				goto retry;
				break;
			case 2:
				skipped++;
				continue;
				break;
			case 3:
			default:
				dfui_progress_free(pr);
				goto dismount;
				break;
			}
		}

		if (!dfui_be_progress_end(a->c))
			abort_backend();
		dfui_progress_free(pr);
	}

	/*
	 * If we were not able to install some distributions let
	 * the user know that it may cause problems, but don't
	 * fail the installation.
	 * Note: the installation may fail further on for other reasons.
	 */
	if (skipped != 0)
		inform(a->c, _("One or more distributions failed to install "
		    "properly. While this may not be a fatal error it may "
		    "cause problems further along in the installation."));

	/*
	 * Write new fstab.
	 */
	snprintf(fstab, sizeof(fstab)/sizeof(char), "%s/etc/fstab",
	    DEFAULT_DESTDIR);
	ff = fopen(fstab, "w");
	if (ff == NULL) {
		inform(a->c, _("Internal Error (%d) - unable to open %s "
		    "for writing"), errno, fstab);
		goto dismount;
	}
	fprintf(ff, "# Device\t\tMountpoint\tFStype\tOptions\t\tDump\tPass#\n");
	for (sp = slice_subpartition_first(storage_get_selected_slice(a->s));
	sp != NULL; sp = subpartition_next(sp)) {
		if (strcmp(subpartition_get_mountpoint(sp), "/") == 0) {
			fprintf(ff, "/dev/%s\t\t%s\t\tufs\trw\t\t1\t1\n",
			    subpartition_get_device_name(sp),
			    subpartition_get_mountpoint(sp));
		} else if (strcmp(subpartition_get_mountpoint(sp), "swap") == 0) {
			fprintf(ff, "/dev/%s\t\tnone\t\tswap\tsw\t\t0\t0\n",
			    subpartition_get_device_name(sp));
		} else if (subpartition_is_mfsbacked(sp)) {
			mfsbacked_size =
			slice_get_capacity(storage_get_selected_slice(a->s)) *
			2048;
			fprintf(ff, "swap\t\t%s\t\t\tmfs\trw,-s%ld\t\t1\t1\n",
			    subpartition_get_mountpoint(sp), mfsbacked_size);
		} else {
			fprintf(ff, "/dev/%s\t\t%s\t\tufs\trw\t\t2\t2\n",
			    subpartition_get_device_name(sp),
			    subpartition_get_mountpoint(sp));
		}
	}
	fclose(ff);

	/*
	 * Mount devfs(5).
	 */
	iovDevfs[0].iov_base = "fstype";
	iovDevfs[0].iov_len = strlen(iovDevfs[0].iov_base) + 1;
	iovDevfs[1].iov_base = "devfs";
	iovDevfs[1].iov_len = strlen(iovDevfs[1].iov_base) + 1;
	iovDevfs[2].iov_base = "fspath";
	iovDevfs[2].iov_len = strlen(iovDevfs[2].iov_base) + 1;
	snprintf(devfsPath, sizeof(devfsPath), "%smnt/dev", a->os_root);
	iovDevfs[3].iov_base = devfsPath;
	iovDevfs[3].iov_len = strlen(iovDevfs[3].iov_base) + 1;
	error = nmount(iovDevfs, 4, 0);
	if (error != 0) {
		inform(a->c, _("Internal Error - nmount of devfs on %s failed "
		    "(%d)."), devfsPath, errno);
		goto dismount;
	}

	fpid = fork();
	if (fpid == 0) {
		struct commands *fcmds;
		int ferr;

		ferr = 0;
		fcmds = commands_new();
		if (chdir(DEFAULT_DESTDIR) != 0)
			i_log(a, "Failed to chdir to %s (%d)",
			    DEFAULT_DESTDIR, errno);
		if (chroot(DEFAULT_DESTDIR) == 0) {
			i_log(a, "chroot'ed to newly installed system\n");
			/*
			* These were obtained from the original sysinstall.
			* We should probably check to make sure they are still needed.
			*/
			mkdir("/var/db", S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
			i_log(a, "%d = mkdir(%s)\t\t errno = %d",
			    error, "var/db", errno);
			close (open("/var/db/mounttab", O_CREAT | O_TRUNC | O_WRONLY,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
			mkdir("/usr/compat",
			S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
			i_log(a, "%d = mkdir(%s)\t\t errno = %d",
			    error, "/usr/compat", errno);
			symlink("/usr/compat", "/compat");
			command_add(fcmds, "/%s", NEWALIASES);
			command_add(fcmds, "/%s -deU -f /etc/mtree/BSD.root.dist -p /",
			MTREE);
			command_add(fcmds, "/%s -deU -f /etc/mtree/BSD.var.dist -p /var",
			MTREE);
			command_add(fcmds, "/%s -deU -f /etc/mtree/BSD.usr.dist -p /usr",
			MTREE);
			command_add(fcmds, "/%s -rf /boot/kernel", RM);
			command_add(fcmds, "/%s /boot/GENERIC /boot/kernel",
			    MV);
			ferr = commands_execute(a, fcmds);
		}
		commands_free(fcmds);
		exit(!ferr);
	}

	/*
	 * Wait for our chrooted process to finish.
	 */
	fstatus = 0;
	if ((wait4(fpid, &fstatus, 0, NULL) < 1) || fstatus != 0) {
		inform(a->c, _("Error - Unable to complete "
		    "post-install fixups."));
		a->result = 0;
	} else
		a->result = 1;


	/* Unmount chroot'ed devfs(5) filesystem. */
	unmount(devfsPath, MNT_FORCE);

	cmds = commands_new();

	/*
	 * Clean up.  In case some file didn't make it, use rm -f
	 */
	cmd = command_add(cmds, "%s%s -f %smnt/tmp/install.log",
	    a->os_root, RM, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);

	/*
	 * Backup the log. This is not terribly important so we don't
	 * need to inform the user of anything if it fails. It will be
	 * in the log output anyway.
	 */
	cmd = command_add(cmds, "%s%s %sinstall.log %smnt/var/log/install.log",
	    a->os_root, CP,
	    a->tmp, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);
	cmd = command_add(cmds, "%s%s 600 %smnt/var/log/install.log",
	    a->os_root, CHMOD, a->os_root);
	command_set_failure_mode(cmd, COMMAND_FAILURE_IGNORE);
	commands_execute(a, cmds);
	commands_free(cmds);

dismount:
	/*
	 * Unmount everything we mounted.  This is done in a seperate
	 * command chain so that partitions are unmounted even if an error
	 * occurs in one of the preceding commands or it is cancelled.
	 */
	if (cdmounted) {
		error = unmount(DIST_MNT, MNT_FORCE);
		i_log(a, "%d = unmount(%s)\t\t errno = %d\n",
			error, DIST_MNT, errno);
	}
	for (sp = slice_subpartition_last(storage_get_selected_slice(a->s));
	     sp != NULL; sp = subpartition_prev(sp)) {
		if (subpartition_is_swap(sp)) {
			snprintf(dev, sizeof(dev) / sizeof(char), "/dev/%s",
			    subpartition_get_device_name(sp));
			error = swapoff(dev);
			i_log(a, "%d = swapoff(%s)\t\t errno = %d\n",
			    error, dev, errno);
			continue;
		}
		snprintf(mnt, sizeof(mnt) / sizeof(char), "%smnt%s",
		    a->os_root, subpartition_get_mountpoint(sp));
		if (unmount(mnt, MNT_FORCE) == -1)
			inform(a->c, _("Warning: subpartition %s was not "
			    "correctly unmounted. Error: %d"), mnt, errno);
	}
}
